import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.svetovid.io.SvetovidReader;
import org.svetovid.io.SvetovidWriter;

/**
 * Format:
 *   /-- desni
 * -(id) vrednost
 *   \-- levi
 */
public class StabloIOPretty implements StabloIO {

    protected static final String EMPTY_SYMBOL = " ";
    protected static final String RIGHT_SYMBOL = "/";
    protected static final String VERTICAL_SYMBOL = "|";
    protected static final String LEFT_SYMBOL = "\\";
    protected static final String HORIZONTAL_SYMBOL = "-";

    protected String nullSymbol;
    protected boolean separated;
    protected int length;

    public StabloIOPretty() {
        this(null, false, 7);
    }

    public StabloIOPretty(String nullSymbol, boolean separated, int length) {
        this.nullSymbol = nullSymbol;
        this.separated = separated;
        this.length = length;
    }

    public String getNullSymbol() {
        return nullSymbol;
    }

    public void setNullSymbol(String nullSymbol) {
        this.nullSymbol = nullSymbol;
    }

    public boolean isSeparated() {
        return separated;
    }

    public void setSeparated(boolean separated) {
        this.separated = separated;
    }

    public int getLength() {
        return length;
    }

    public void setLength(int length) {
        if (length < 3) {
            throw new IllegalArgumentException("length");
        }
        this.length = length;
    }

    @Override
    public Stablo readStablo(SvetovidReader in) {
        return parseStablo(in, nullSymbol, length);
    }

    protected Stablo parseStablo(SvetovidReader in, String nullSymbol, int length) {
        List<Stablo> elements = new ArrayList<>();
        List<Integer> levels = new ArrayList<>();
        Pattern levelPattern = Pattern.compile("[\\Q" + LEFT_SYMBOL + HORIZONTAL_SYMBOL + RIGHT_SYMBOL + "\\E]");
        String line = in.readLine();
        while ((line != null) && !line.isEmpty()) {
            Matcher matcher = levelPattern.matcher(line);
            int level = -1;
            if (matcher.find()) {
                level = matcher.start();
            }
            if (level != -1 && (nullSymbol == null || !line.endsWith(nullSymbol))) {
                Stablo stablo = parseStablo(line);
                elements.add(stablo);
                levels.add(level);
            }
            line = in.readLine();
        }
        Stablo stablo = formStablo(0, elements.size(), levels, elements);
        return stablo;
    }

    private Stablo parseStablo(String line) {
        int id = -1;
        String vrednost = null;
        int beginIndex = line.indexOf('(');
        int endIndex = line.indexOf(')');
        if ((beginIndex != -1) && (endIndex != -1) && (beginIndex < endIndex)) {
            vrednost = line.substring(beginIndex + 1, endIndex);
            try {
                id = Integer.parseInt(vrednost);
            } catch (NumberFormatException e) {
                throw new NumberFormatException(line);
            }
            vrednost = line.substring(endIndex + 2);
        } else {
            throw new NumberFormatException(line);
        }
        Stablo stablo = new Stablo(id, vrednost);
        return stablo;
    }

    private Stablo formStablo(int beginIndex, int endIndex, List<Integer> levels, List<Stablo> elements) {
        if (beginIndex >= endIndex) {
            return null;
        }
        int minIndex = beginIndex;
        int minLevel = levels.get(minIndex);
        for (int i = beginIndex + 1; i < endIndex; i++) {
            int level = levels.get(i);
            if (level < minLevel) {
                minLevel = level;
                minIndex = i;
            }
        }
        Stablo stablo = elements.get(minIndex);
        Stablo levi = formStablo(minIndex + 1, endIndex, levels, elements);
        Stablo desni = formStablo(beginIndex, minIndex, levels, elements);
        stablo.levi = levi;
        stablo.desni = desni;
        return stablo;
    }

    @Override
    public void printStablo(SvetovidWriter out, Stablo stablo) {
        StringBuilder builder = new StringBuilder();
        appendTree(builder, stablo, nullSymbol, separated, length);
        out.print(builder.toString());
    }

    protected void appendTree(StringBuilder builder, Stablo stablo, String nullSymbol, boolean separated, int length) {
        String[] buildingBlocks = generateBuildingBlocks(length);
        appendRight(builder, stablo, nullSymbol, separated, buildingBlocks, true, buildingBlocks[5]);
        appendNode(builder, stablo, nullSymbol != null ? nullSymbol : "|", buildingBlocks[4]);
        appendLeft(builder, stablo, nullSymbol, separated, buildingBlocks, false, buildingBlocks[5]);
    }

    protected void appendNode(StringBuilder builder, Stablo stablo, String nullSymbol, String prefix) {
        builder.append(prefix);
        if (stablo == null) {
            builder.append(nullSymbol);
        } else {
            builder.append("(");
            builder.append(stablo.id);
            builder.append(") ");
            builder.append(stablo.vrednost);
        }
        builder.append("\n");
    }

    protected void appendRight(StringBuilder builder, Stablo stablo, String nullSymbol, boolean separated, String[] buildingBlocks, boolean isRight, String prefix) {
        if (stablo == null) {
            return;
        }
        if ((nullSymbol != null) || (stablo.desni != null)) {
            appendSubtree(builder, stablo.desni, nullSymbol, separated, buildingBlocks, true, prefix);
            if (separated) {
                appendEmpty(builder, buildingBlocks, prefix);
            }
        }
    }

    protected void appendLeft(StringBuilder builder, Stablo stablo, String nullSymbol, boolean separated, String[] buildingBlocks, boolean isRight, String prefix) {
        if (stablo == null) {
            return;
        }
        if ((nullSymbol != null) || (stablo.levi != null)) {
            if (separated) {
                appendEmpty(builder, buildingBlocks, prefix);
            }
            appendSubtree(builder, stablo.levi, nullSymbol, separated, buildingBlocks, false, prefix);
        }
    }

    protected void appendEmpty(StringBuilder builder, String[] buildingBlocks, String prefix) {
        builder.append(prefix);
        builder.append(buildingBlocks[2]);
        builder.append("\n");
    }

    protected void appendSubtree(StringBuilder builder, Stablo stablo, String nullSymbol, boolean separated, String[] buildingBlocks, boolean isRight, String prefix) {
        String mojPrefix = prefix;
        if (isRight == true) {
            mojPrefix = mojPrefix + buildingBlocks[1];
        }
        if (isRight == false) {
            mojPrefix = mojPrefix + buildingBlocks[3];
        }
        String noviPrefix = prefix + (!isRight ? buildingBlocks[2] : buildingBlocks[0]);
        appendRight(builder, stablo, nullSymbol, separated, buildingBlocks, isRight, noviPrefix);
        appendNode(builder, stablo, nullSymbol, mojPrefix);
        noviPrefix = prefix + (isRight ? buildingBlocks[2] : buildingBlocks[0]);
        appendLeft(builder, stablo, nullSymbol, separated, buildingBlocks, isRight, noviPrefix);
    }

    private String[] generateBuildingBlocks(int length) {
        String[] blocks = new String[6];
        blocks[0] = generateBlock(EMPTY_SYMBOL, EMPTY_SYMBOL, EMPTY_SYMBOL, length - 2);
        blocks[1] = generateBlock(EMPTY_SYMBOL, RIGHT_SYMBOL, HORIZONTAL_SYMBOL, length - 2);
        blocks[2] = generateBlock(EMPTY_SYMBOL, VERTICAL_SYMBOL, EMPTY_SYMBOL, length - 2);
        blocks[3] = generateBlock(EMPTY_SYMBOL, LEFT_SYMBOL, HORIZONTAL_SYMBOL, length - 2);
        blocks[4] = HORIZONTAL_SYMBOL;
        blocks[5] = EMPTY_SYMBOL;
        return blocks;
    }

    protected String generateBlock(String emptySymbol, String startSymbol, String repeatSymbol, int repeatCount) {
        StringBuilder builder = new StringBuilder();
        builder.append(emptySymbol);
        builder.append(startSymbol);
        for (int i = 0; i < repeatCount; i++) {
            builder.append(repeatSymbol);
        }
        return builder.toString();
    }
}
